from abc import ABC, abstractmethod
from typing import NoReturn

from discord import Interaction
from discord.ext.commands import Context


class AbstractCommandErrorHandler(ABC):
    """An abstract command error handler."""

    @abstractmethod
    async def should_handle_error(self, error: Exception) -> bool:
        """A predicate that determines whether the error should be handled or not."""
        raise NotImplementedError

    @abstractmethod
    async def handle_app_command_error(self, interaction: Interaction, error: Exception) -> NoReturn:
        """Handle error raised in the context of app commands."""
        raise NotImplementedError

    @abstractmethod
    async def handle_text_command_error(self, context: Context, error: Exception) -> NoReturn:
        """Handle error raised in the context of text commands."""
        raise NotImplementedError



"""Utilities related to custom caches."""

import functools
import typing
from collections import OrderedDict


class AsyncCache:
    """
    LRU cache implementation for coroutines.

    Once the cache exceeds the maximum size, keys are deleted in FIFO order.

    An offset may be optionally provided to be applied to the coroutine's arguments when creating the cache key.
    """

    def __init__(self, max_size: int = 128):
        """
        Initialise a new :obj:`AsyncCache` instance.

        Args:
            max_size: How many items to store in the cache.
        """
        self._cache = OrderedDict()
        self._max_size = max_size

    def __call__(self, arg_offset: int = 0) -> typing.Callable:
        """
        Decorator for async cache.

        Args:
            arg_offset: The offset for the position of the key argument.

        Returns:
            A decorator to wrap the target function.
        """

        def decorator(function: typing.Callable) -> typing.Callable:
            """
            Define the async cache decorator.

            Args:
                function: The function to wrap.

            Returns:
                The wrapped function.
            """

            @functools.wraps(function)
            async def wrapper(*args) -> typing.Any:
                """Decorator wrapper for the caching logic."""
                key = args[arg_offset:]

                if key not in self._cache:
                    if len(self._cache) > self._max_size:
                        self._cache.popitem(last=False)

                    self._cache[key] = await function(*args)
                return self._cache[key]
            return wrapper
        return decorator

    def clear(self) -> None:
        """Clear cache instance."""
        self._cache.clear()



"""Useful helper functions for interacting with various discord channel objects."""

import discord
from discord.ext.commands import Bot

from pydis_core.utils import logging

log = logging.get_logger(__name__)


def is_in_category(channel: discord.TextChannel, category_id: int) -> bool:
    """
    Return whether the given ``channel`` in the the category with the id ``category_id``.

    Args:
        channel: The channel to check.
        category_id: The category to check for.

    Returns:
        A bool depending on whether the channel is in the category.
    """
    return getattr(channel, "category_id", None) == category_id


async def get_or_fetch_channel(
    bot: Bot,
    channel_id: int,
) -> discord.abc.GuildChannel | discord.abc.PrivateChannel | discord.Thread:
    """
    Attempt to get or fetch the given ``channel_id`` from the bots cache, and return it.

    Args:
        bot: The :obj:`discord.ext.commands.Bot` instance to use for getting/fetching.
        channel_id: The channel to get/fetch.

    Raises:
        :exc:`discord.InvalidData`
            An unknown channel type was received from Discord.
        :exc:`discord.HTTPException`
            Retrieving the channel failed.
        :exc:`discord.NotFound`
            Invalid Channel ID.
        :exc:`discord.Forbidden`
            You do not have permission to fetch this channel.

    Returns:
        The channel from the ID.
    """
    log.trace(f"Getting the channel {channel_id}.")

    channel = bot.get_channel(channel_id)
    if not channel:
        log.debug(f"Channel {channel_id} is not in cache; fetching from API.")
        channel = await bot.fetch_channel(channel_id)

    log.trace(f"Channel #{channel} ({channel_id}) retrieved.")
    return channel



from collections.abc import Callable
from typing import NoReturn

from discord import Interaction
from discord.ext.commands import Context

from pydis_core.utils.error_handling.commands import AbstractCommandErrorHandler
from pydis_core.utils.logging import get_logger

log = get_logger(__name__)


class CommandErrorManager:
    """A class that registers error handlers and handles all command related errors."""

    def __init__(self, default: AbstractCommandErrorHandler, handlers: list[AbstractCommandErrorHandler] | None = None):
        self.handlers = handlers or []
        self.registered_handlers = set(type(handler).__name__ for handler in self.handlers)
        self.default = default

    async def handle_error(
        self,
        error: Exception,
        context_or_interaction: Context | Interaction
    ) -> None:
        """
        Handle a discord exception.

        This works simply by iterating over the available handlers, and chooses the one capable of handling the error.

        Instead of having some priority system to distinguish the execution order of handlers, we will
        let the `should_handle_error` do the job.

        The default needs to always be last, since the container can either be initialized with the handlers, or
        we can add handlers using the `register_handler` function that we don't know when it will be invoked, if ever.
        """
        for handler in self.handlers + [self.default]:
            if await handler.should_handle_error(error):
                callback = self._determine_callback(handler, context_or_interaction)
                await callback(context_or_interaction, error)
                break

    def register_handler(self, handler: AbstractCommandErrorHandler) -> None:
        """Register a command error handler."""
        handler_name = type(handler).__name__
        if handler_name in self.registered_handlers:
            log.info(f"Skipping registration of command error handler '{handler_name}' as it's already registered.")
            return

        self.handlers.append(handler)
        self.registered_handlers.add(type(handler).__name__)

    @staticmethod
    def _determine_callback(
        handler: AbstractCommandErrorHandler,
        context_or_interaction: Context | Interaction
    ) -> Callable[[Context, Exception], NoReturn] | Callable[[Interaction, Exception], NoReturn] | None:
        """Determines the handling callback that will be used."""
        if isinstance(context_or_interaction, Context):
            return handler.handle_text_command_error
        return handler.handle_app_command_error



"""Contains all common monkey patches, used to alter discord to fit our needs."""

import logging
from datetime import datetime, timedelta, timezone
from functools import partial, partialmethod

from discord import Forbidden, http
from discord.ext import commands

log = logging.getLogger(__name__)


class _Command(commands.Command):
    """
    A :obj:`discord.ext.commands.Command` subclass which supports root aliases.

    A ``root_aliases`` keyword argument is added, which is a sequence of alias names that will act as
    top-level commands rather than being aliases of the command's group. It's stored as an attribute
    also named ``root_aliases``.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.root_aliases = kwargs.get("root_aliases", [])

        if not isinstance(self.root_aliases, list | tuple):
            raise TypeError("Root aliases of a command must be a list or a tuple of strings.")


class _Group(commands.Group, _Command):
    """
    A :obj:`discord.ext.commands.Group` subclass which supports root aliases.

    A ``root_aliases`` keyword argument is added, which is a sequence of alias names that will act as
    top-level groups rather than being aliases of the command's group. It's stored as an attribute
    also named ``root_aliases``.
    """


def _patch_typing() -> None:
    """
    Sometimes Discord turns off typing events by throwing 403s.

    Handle those issues by patching discord's internal ``send_typing`` method so it ignores 403s in general.
    """
    log.debug("Patching send_typing, which should fix things breaking when Discord disables typing events. Stay safe!")

    original = http.HTTPClient.send_typing
    last_403: datetime | None = None

    async def honeybadger_type(self: http.HTTPClient, channel_id: int) -> None:
        nonlocal last_403
        if last_403 and (datetime.now(tz=timezone.utc) - last_403) < timedelta(minutes=5):
            log.warning("Not sending typing event, we got a 403 less than 5 minutes ago.")
            return
        try:
            await original(self, channel_id)
        except Forbidden:
            last_403 = datetime.now(tz=timezone.utc)
            log.warning("Got a 403 from typing event!")

    http.HTTPClient.send_typing = honeybadger_type


def _apply_monkey_patches() -> None:
    """This is surfaced directly in pydis_core.utils.apply_monkey_patches()."""
    commands.command = partial(commands.command, cls=_Command)
    commands.GroupMixin.command = partialmethod(commands.GroupMixin.command, cls=_Command)

    commands.group = partial(commands.group, cls=_Group)
    commands.GroupMixin.group = partialmethod(commands.GroupMixin.group, cls=_Group)
    _patch_typing()




from discord import Message
from discord.ext.commands import BadArgument, Context, clean_content


async def clean_text_or_reply(ctx: Context, text: str | None = None) -> str:
    """
    Cleans a text argument or replied message's content.

    Args:
        ctx: The command's context
        text: The provided text argument of the command (if given)

    Raises:
        :exc:`discord.ext.commands.BadArgument`
            `text` wasn't provided and there's no reply message / reply message content.

    Returns:
         The cleaned version of `text`, if given, else replied message.
    """
    clean_content_converter = clean_content(fix_channel_mentions=True)

    if text:
        return await clean_content_converter.convert(ctx, text)

    if (
        (replied_message := getattr(ctx.message.reference, "resolved", None))  # message has a cached reference
        and isinstance(replied_message, Message)  # referenced message hasn't been deleted
    ):
        if not (content := ctx.message.reference.resolved.content):
            # The referenced message doesn't have a content (e.g. embed/image), so raise error
            raise BadArgument("The referenced message doesn't have a text content.")

        return await clean_content_converter.convert(ctx, content)

    # No text provided, and either no message was referenced or we can't access the content
    raise BadArgument("Couldn't find text to clean. Provide a string or reply to a message to use its content.")



from discord import Forbidden, Message

from pydis_core.utils import logging

log = logging.get_logger(__name__)


async def handle_forbidden_from_block(error: Forbidden, message: Message | None = None) -> None:
    """
    Handles ``discord.Forbidden`` 90001 errors, or re-raises if ``error`` isn't a 90001 error.

    Args:
        error: The raised ``discord.Forbidden`` to check.
        message: The message to reply to and include in logs, if error is 90001 and message is provided.
    """
    if error.code != 90001:
        # The error ISN'T caused by the bot attempting to add a reaction
        # to a message whose author has blocked the bot, so re-raise it
        raise error

    if not message:
        log.info("Failed to add reaction(s) to a message since the message author has blocked the bot")
        return

    if message:
        log.info(
            "Failed to add reaction(s) to message %d-%d since the message author (%d) has blocked the bot",
            message.channel.id,
            message.id,
            message.author.id,
        )
        await message.channel.send(
            f":x: {message.author.mention} failed to add reaction(s) to your message as you've blocked me.",
            delete_after=30
        )



"""Utils for manipulating functions."""

from __future__ import annotations

import functools
import inspect
import types
import typing
from collections.abc import Callable, Sequence, Set

__all__ = [
    "GlobalNameConflictError",
    "command_wraps",
    "get_arg_value",
    "get_arg_value_wrapper",
    "get_bound_args",
    "update_wrapper_globals",
]


if typing.TYPE_CHECKING:
    _P = typing.ParamSpec("_P")
    _R = typing.TypeVar("_R")

Argument = int | str
BoundArgs = typing.OrderedDict[str, typing.Any]
Decorator = typing.Callable[[typing.Callable], typing.Callable]
ArgValGetter = typing.Callable[[BoundArgs], typing.Any]


class GlobalNameConflictError(Exception):
    """Raised on a conflict between the globals used to resolve annotations of a wrapped function and its wrapper."""


def get_arg_value(name_or_pos: Argument, arguments: BoundArgs) -> typing.Any:
    """
    Return a value from `arguments` based on a name or position.

    Arguments:
        arguments: An ordered mapping of parameter names to argument values.
    Returns:
        Value from `arguments` based on a name or position.
    Raises:
        TypeError: `name_or_pos` isn't a str or int.
        ValueError: `name_or_pos` does not match any argument.
    """
    if isinstance(name_or_pos, int):
        # Convert arguments to a tuple to make them indexable.
        arg_values = tuple(arguments.items())
        arg_pos = name_or_pos

        try:
            _name, value = arg_values[arg_pos]
            return value
        except IndexError:
            raise ValueError(f"Argument position {arg_pos} is out of bounds.")
    elif isinstance(name_or_pos, str):
        arg_name = name_or_pos
        try:
            return arguments[arg_name]
        except KeyError:
            raise ValueError(f"Argument {arg_name!r} doesn't exist.")
    else:
        raise TypeError("'arg' must either be an int (positional index) or a str (keyword).")


def get_arg_value_wrapper(
    decorator_func: typing.Callable[[ArgValGetter], Decorator],
    name_or_pos: Argument,
    func: typing.Callable[[typing.Any], typing.Any] | None = None,
) -> Decorator:
    """
    Call `decorator_func` with the value of the arg at the given name/position.

    Arguments:
        decorator_func: A function that must accept a callable as a parameter to which it will pass a mapping of
            parameter names to argument values of the function it's decorating.
        name_or_pos: The name/position of the arg to get the value from.
        func: An optional callable which will return a new value given the argument's value.

    Returns:
        The decorator returned by `decorator_func`.
    """
    def wrapper(args: BoundArgs) -> typing.Any:
        value = get_arg_value(name_or_pos, args)
        if func:
            value = func(value)
        return value

    return decorator_func(wrapper)


def get_bound_args(func: typing.Callable, args: tuple, kwargs: dict[str, typing.Any]) -> BoundArgs:
    """
    Bind `args` and `kwargs` to `func` and return a mapping of parameter names to argument values.

    Default parameter values are also set.

    Args:
        args: The arguments to bind to ``func``
        kwargs: The keyword arguments to bind to ``func``
        func: The function to bind ``args`` and ``kwargs`` to
    Returns:
        A mapping of parameter names to argument values.
    """
    sig = inspect.signature(func)
    bound_args = sig.bind(*args, **kwargs)
    bound_args.apply_defaults()

    return bound_args.arguments


def update_wrapper_globals(
    wrapper: Callable[_P, _R],
    wrapped: Callable[_P, _R],
    *,
    ignored_conflict_names: Set[str] = frozenset(),
) -> Callable[_P, _R]:
    r"""
    Create a copy of ``wrapper``\, the copy's globals are updated with ``wrapped``\'s globals.

    For forwardrefs in command annotations, discord.py uses the ``__global__`` attribute of the function
    to resolve their values. This breaks for decorators that replace the function because they have
    their own globals.

    .. warning::
        This function captures the state of ``wrapped``\'s module's globals when it's called;
        changes won't be reflected in the new function's globals.

    Args:
        wrapper: The function to wrap.
        wrapped: The function to wrap with.
        ignored_conflict_names: A set of names to ignore if a conflict between them is found.

    Raises:
        :exc:`GlobalNameConflictError`:
            If ``wrapper`` and ``wrapped`` share a global name that's also used in ``wrapped``\'s typehints,
            and is not in ``ignored_conflict_names``.
    """
    wrapped = typing.cast(types.FunctionType, wrapped)
    wrapper = typing.cast(types.FunctionType, wrapper)

    annotation_global_names = (
        ann.split(".", maxsplit=1)[0] for ann in wrapped.__annotations__.values() if isinstance(ann, str)
    )
    # Conflicting globals from both functions' modules that are also used in the wrapper and in wrapped's annotations.
    shared_globals = (
        set(wrapper.__code__.co_names)
        & set(annotation_global_names)
        & set(wrapped.__globals__)
        & set(wrapper.__globals__)
        - ignored_conflict_names
    )
    if shared_globals:
        raise GlobalNameConflictError(
            f"wrapper and the wrapped function share the following "
            f"global names used by annotations: {', '.join(shared_globals)}. Resolve the conflicts or add "
            f"the name to the `ignored_conflict_names` set to suppress this error if this is intentional."
        )

    new_globals = wrapper.__globals__.copy()
    new_globals.update((k, v) for k, v in wrapped.__globals__.items() if k not in wrapper.__code__.co_names)
    return types.FunctionType(
        code=wrapper.__code__,
        globals=new_globals,
        name=wrapper.__name__,
        argdefs=wrapper.__defaults__,
        closure=wrapper.__closure__,
    )


def command_wraps(
    wrapped: Callable[_P, _R],
    assigned: Sequence[str] = functools.WRAPPER_ASSIGNMENTS,
    updated: Sequence[str] = functools.WRAPPER_UPDATES,
    *,
    ignored_conflict_names: Set[str] = frozenset(),
) -> Callable[[Callable[_P, _R]], Callable[_P, _R]]:
    r"""
    Update the decorated function to look like ``wrapped``\, and update globals for discord.py forwardref evaluation.

    See :func:`update_wrapper_globals` for more details on how the globals are updated.

    Args:
        wrapped: The function to wrap with.
        assigned: Sequence of attribute names that are directly assigned from ``wrapped`` to ``wrapper``.
        updated: Sequence of attribute names that are ``.update``d on ``wrapper`` from the attributes on ``wrapped``.
        ignored_conflict_names: A set of names to ignore if a conflict between them is found.

    Returns:
        A decorator that behaves like :func:`functools.wraps`,
        with the wrapper replaced with the function :func:`update_wrapper_globals` returned.
    """
    def decorator(wrapper: Callable[_P, _R]) -> Callable[_P, _R]:
        return functools.update_wrapper(
            update_wrapper_globals(wrapper, wrapped, ignored_conflict_names=ignored_conflict_names),
            wrapped,
            assigned,
            updated,
        )

    return decorator



"""An API wrapper around the Site API."""

import asyncio
from urllib.parse import quote as quote_url

import aiohttp

from pydis_core.utils.logging import get_logger

log = get_logger(__name__)


class ResponseCodeError(ValueError):
    """Raised in :meth:`APIClient.request` when a non-OK HTTP response is received."""

    def __init__(
        self,
        response: aiohttp.ClientResponse,
        response_json: dict | None = None,
        response_text: str | None = None
    ):
        """
        Initialize a new :obj:`ResponseCodeError` instance.

        Args:
            response (:obj:`aiohttp.ClientResponse`): The response object from the request.
            response_json: The JSON response returned from the request, if any.
            response_text: The text of the request, if any.
        """
        self.status = response.status
        self.response_json = response_json or {}
        self.response_text = response_text
        self.response = response

    def __str__(self):
        """Return a string representation of the error."""
        response = self.response_json or self.response_text
        return f"Status: {self.status} Response: {response}"


class APIClient:
    """A wrapper for the Django Site API."""

    session: aiohttp.ClientSession | None = None
    loop: asyncio.AbstractEventLoop = None

    def __init__(self, site_api_url: str, site_api_token: str, **session_kwargs):
        """
        Initialize a new :obj:`APIClient` instance.

        Args:
            site_api_url: The URL of the site API.
            site_api_token: The token to use for authentication.
            session_kwargs: Keyword arguments to pass to the :obj:`aiohttp.ClientSession` constructor.
        """
        self.site_api_url = site_api_url

        auth_headers = {
            "Authorization": f"Token {site_api_token}"
        }

        if "headers" in session_kwargs:
            session_kwargs["headers"].update(auth_headers)
        else:
            session_kwargs["headers"] = auth_headers

        # aiohttp will complain if APIClient gets instantiated outside a coroutine. Thankfully, we
        # don't and shouldn't need to do that, so we can avoid scheduling a task to create it.
        self.session = aiohttp.ClientSession(**session_kwargs)

    def _url_for(self, endpoint: str) -> str:
        return f"{self.site_api_url}/{quote_url(endpoint)}"

    async def close(self) -> None:
        """Close the aiohttp session."""
        await self.session.close()

    @staticmethod
    async def maybe_raise_for_status(response: aiohttp.ClientResponse, should_raise: bool) -> None:
        """
        Raise :exc:`ResponseCodeError` for non-OK response if an exception should be raised.

        Args:
            response (:obj:`aiohttp.ClientResponse`): The response to check.
            should_raise: Whether or not to raise an exception.

        Raises:
            :exc:`ResponseCodeError`:
                If the response is not OK and ``should_raise`` is True.
        """
        if should_raise and response.status >= 400:
            try:
                response_json = await response.json()
                raise ResponseCodeError(response=response, response_json=response_json)
            except aiohttp.ContentTypeError:
                response_text = await response.text()
                raise ResponseCodeError(response=response, response_text=response_text)

    async def request(self, method: str, endpoint: str, *, raise_for_status: bool = True, **kwargs) -> dict | None:
        """
        Send an HTTP request to the site API and return the JSON response.

        Args:
            method: The HTTP method to use.
            endpoint: The endpoint to send the request to.
            raise_for_status: Whether or not to raise an exception if the response is not OK.
            **kwargs: Any extra keyword arguments to pass to :func:`aiohttp.request`.

        Returns:
            The JSON response the API returns, or :obj:`None` if the response code is 204.

        Raises:
            :exc:`ResponseCodeError`:
                If the response is not OK and ``raise_for_status`` is True.
        """
        async with self.session.request(method.upper(), self._url_for(endpoint), **kwargs) as resp:
            if resp.status == 204:
                return None

            await self.maybe_raise_for_status(resp, raise_for_status)
            return await resp.json()

    async def get(self, endpoint: str, *, raise_for_status: bool = True, **kwargs) -> dict | None:
        """Equivalent to :meth:`APIClient.request` with GET passed as the method."""
        return await self.request("GET", endpoint, raise_for_status=raise_for_status, **kwargs)

    async def patch(self, endpoint: str, *, raise_for_status: bool = True, **kwargs) -> dict | None:
        """Equivalent to :meth:`APIClient.request` with PATCH passed as the method."""
        return await self.request("PATCH", endpoint, raise_for_status=raise_for_status, **kwargs)

    async def post(self, endpoint: str, *, raise_for_status: bool = True, **kwargs) -> dict | None:
        """Equivalent to :meth:`APIClient.request` with POST passed as the method."""
        return await self.request("POST", endpoint, raise_for_status=raise_for_status, **kwargs)

    async def put(self, endpoint: str, *, raise_for_status: bool = True, **kwargs) -> dict | None:
        """Equivalent to :meth:`APIClient.request` with PUT passed as the method."""
        return await self.request("PUT", endpoint, raise_for_status=raise_for_status, **kwargs)

    async def delete(self, endpoint: str, *, raise_for_status: bool = True, **kwargs) -> dict | None:
        """Equivalent to :meth:`APIClient.request` with DELETE passed as the method."""
        return await self.request("DELETE", endpoint, raise_for_status=raise_for_status, **kwargs)


__all__ = ["APIClient", "ResponseCodeError"]



from collections.abc import Sequence
from typing import Literal

from discord import ButtonStyle, HTTPException, Interaction, Member, Message, NotFound, User, ui

from pydis_core.utils.logging import get_logger
from pydis_core.utils.scheduling import create_task

log = get_logger(__name__)


def user_has_access(
    user: User | Member,
    *,
    allowed_users: Sequence[int] = (),
    allowed_roles: Sequence[int] = (),
) -> bool:
    """
    Return whether the user is in the allowed_users list, or has a role from allowed_roles.

    Args:
        user: The user to check
        allowed_users: A sequence of user ids that are allowed access
        allowed_roles: A sequence of role ids that are allowed access
    """
    if user.id in allowed_users or any(role.id in allowed_roles for role in getattr(user, "roles", [])):
        return True
    return False


async def _handle_modify_message(message: Message, action: Literal["edit", "delete"]) -> None:
    """Remove the view from, or delete the given message depending on the specified action."""
    try:
        if action == "edit":
            await message.edit(view=None)
        elif action == "delete":
            await message.delete()
    except HTTPException as e:
        # Cover the case where this message has been deleted by external means,
        # or the message is now in an archived/locked thread.
        if e.code == 50083:
            log.debug(f"Could not {action} message {message.id} due to it being in an archived thread.")
        elif isinstance(e, NotFound):
            log.info(f"Could not find message {message.id} when attempting to {action} it.")
        else:
            log.error(f"Could not {action} message {message.id} due to Discord HTTP error:\n{e!s}")


class ViewWithUserAndRoleCheck(ui.View):
    """
    A view that allows the original invoker and moderators to interact with it.

    Args:
        allowed_users: A sequence of user's ids who are allowed to interact with the view.
        allowed_roles: A sequence of role ids that are allowed to interact with the view.
        timeout: Timeout in seconds from last interaction with the UI before no longer accepting input.
            If ``None`` then there is no timeout.
        message: The message to remove the view from on timeout. This can also be set with
            ``view.message = await ctx.send( ... )``` , or similar, after the view is instantiated.
    """

    def __init__(
        self,
        *,
        allowed_users: Sequence[int],
        allowed_roles: Sequence[int],
        timeout: float | None = 180.0,
        message: Message | None = None
    ) -> None:
        super().__init__(timeout=timeout)
        self.allowed_users = allowed_users
        self.allowed_roles = allowed_roles
        self.message = message

    async def interaction_check(self, interaction: Interaction) -> bool:
        """
        Ensure the user clicking the button is the view invoker, or a moderator.

        Args:
            interaction: The interaction that occurred.
        """
        if user_has_access(
            interaction.user,
            allowed_users=self.allowed_users,
            allowed_roles=self.allowed_roles,
        ):
            log.trace(
                "Allowed interaction by %s (%d) on %d as they are an allowed user or have an allowed role.",
                interaction.user,
                interaction.user.id,
                interaction.message.id,
            )
            return True

        await interaction.response.send_message("This is not your button to click!", ephemeral=True)
        return False

    def stop(self) -> None:
        """Stop listening for interactions, and remove the view from ``self.message`` if set."""
        super().stop()
        if self.message:
            create_task(_handle_modify_message(self.message, "edit"))

    async def on_timeout(self) -> None:
        """Remove the view from ``self.message`` if set."""
        if self.message:
            await _handle_modify_message(self.message, "edit")


class DeleteMessageButton(ui.Button):
    """
    A button that can be added to a view to delete the message containing the view on click.

    This button itself carries out no interaction checks, these should be done by the parent view.

    See :obj:`pydis_core.utils.interactions.ViewWithUserAndRoleCheck` for a view that implements basic checks.

    Args:
        style (:literal-url:`ButtonStyle <https://discordpy.readthedocs.io/en/latest/interactions/api.html#discord.ButtonStyle>`):
            The style of the button, set to ``ButtonStyle.secondary`` if not specified.
        label: The label of the button, set to "Delete" if not specified.
    """

    def __init__(
        self,
        *,
        style: ButtonStyle = ButtonStyle.secondary,
        label: str = "Delete",
        **kwargs
    ):
        super().__init__(style=style, label=label, **kwargs)

    async def callback(self, interaction: Interaction) -> None:
        """Delete the original message on button click."""
        await _handle_modify_message(interaction.message, "delete")



"""Utilities for loading Discord extensions."""

import importlib
import inspect
import pkgutil
import types
from typing import NoReturn


def unqualify(name: str) -> str:
    """
    Return an unqualified name given a qualified module/package ``name``.

    Args:
        name: The module name to unqualify.

    Returns:
        The unqualified module name.
    """
    return name.rsplit(".", maxsplit=1)[-1]


def ignore_module(module: pkgutil.ModuleInfo) -> bool:
    """Return whether the module with name `name` should be ignored."""
    return any(name.startswith("_") for name in module.name.split("."))


def walk_extensions(module: types.ModuleType) -> frozenset[str]:
    """
    Return all extension names from the given module.

    Args:
        module (types.ModuleType): The module to look for extensions in.

    Returns:
        A set of strings that can be passed directly to :obj:`discord.ext.commands.Bot.load_extension`.
    """

    def on_error(name: str) -> NoReturn:
        raise ImportError(name=name)  # pragma: no cover

    modules = set()

    for module_info in pkgutil.walk_packages(module.__path__, f"{module.__name__}.", onerror=on_error):
        if ignore_module(module_info):
            # Ignore modules/packages that have a name starting with an underscore anywhere in their trees.
            continue

        if module_info.ispkg:
            imported = importlib.import_module(module_info.name)
            if not inspect.isfunction(getattr(imported, "setup", None)):
                # If it lacks a setup function, it's not an extension.
                continue

        modules.add(module_info.name)

    return frozenset(modules)



import datetime
from collections.abc import Callable, Container, Iterable

from discord.ext.commands import (
    BucketType,
    CheckFailure,
    Cog,
    Command,
    CommandOnCooldown,
    Context,
    Cooldown,
    CooldownMapping,
)

from pydis_core.utils.logging import get_logger

log = get_logger(__name__)


class ContextCheckFailure(CheckFailure):
    """Raised when a context-specific check fails."""

    def __init__(self, redirect_channel: int | None) -> None:
        self.redirect_channel = redirect_channel

        if redirect_channel:
            redirect_message = f" here. Please use the <#{redirect_channel}> channel instead"
        else:
            redirect_message = ""

        error_message = f"You are not allowed to use that command{redirect_message}."

        super().__init__(error_message)


class InWhitelistCheckFailure(ContextCheckFailure):
    """Raised when the `in_whitelist` check fails."""


def in_whitelist_check(
    ctx: Context,
    redirect: int,
    channels: Container[int] = (),
    categories: Container[int] = (),
    roles: Container[int] = (),
    fail_silently: bool = False,
) -> bool:
    """
    Check if a command was issued in a context that is whitelisted by channel, category, or roles.

    Args:
        ctx (:obj:`discord.ext.commands.Context`): The context in which the command is being invoked.
        redirect (int | None): The channel ID to redirect the user to, if any.
        channels (Container[int]): Whitelisted channel IDs. Defaults to ().
        categories (Container[int]): Whitelisted category IDs. Defaults to ().
        roles (Container[int]): Whitelisted role IDs. Defaults to ().
        fail_silently (bool): Whether to fail silently without raising an exception. Defaults to False.

    Returns:
        bool: True if the command is used in a whitelisted context; False otherwise.

    Raises:
        InWhitelistCheckFailure: If the context is not whitelisted and `fail_silently` is False.
    """
    if redirect not in channels:
        # It does not make sense for the channel whitelist to not contain the redirection
        # channel (if applicable). That's why we add the redirection channel to the `channels`
        # container if it's not already in it. As we allow any container type to be passed,
        # we first create a tuple in order to safely add the redirection channel.
        #
        # Note: It's possible for the redirect channel to be in a whitelisted category, but
        # there's no easy way to check that and as a channel can easily be moved in and out of
        # categories, it's probably not wise to rely on its category in any case.
        channels = tuple(channels) + (redirect,)

    if channels and ctx.channel.id in channels:
        log.trace(f"{ctx.author} may use the `{ctx.command.name}` command as they are in a whitelisted channel.")
        return True

    # Only check the category id if we have a category whitelist and the channel has a `category_id`
    if categories and hasattr(ctx.channel, "category_id") and ctx.channel.category_id in categories:
        log.trace(f"{ctx.author} may use the `{ctx.command.name}` command as they are in a whitelisted category.")
        return True

    # Only check the roles whitelist if we have one and ensure the author's roles attribute returns
    # an iterable to prevent breakage in DM channels (for if we ever decide to enable commands there).
    if roles and any(r.id in roles for r in getattr(ctx.author, "roles", ())):
        log.trace(f"{ctx.author} may use the `{ctx.command.name}` command as they have a whitelisted role.")
        return True

    log.trace(f"{ctx.author} may not use the `{ctx.command.name}` command within this context.")

    # Some commands are secret, and should produce no feedback at all.
    if not fail_silently:
        raise InWhitelistCheckFailure(redirect)
    return False


def cooldown_with_role_bypass(
    rate: int,
    per: float,
    type: BucketType = BucketType.default,
    *,
    bypass_roles: Iterable[int]
) -> Callable:
    """
    Decorate a command to have a cooldown, which can be bypassed by users with specified roles.

    Note: This replaces the `Command.before_invoke` callback, which *might* introduce problems in the future.

    Args:
        rate (int): Number of times a command can be used before triggering a cooldown.
        per (float): The duration (in seconds) for how long the cooldown lasts.
        type (:obj:`discord.ext.commands.BucketType`): The type of cooldown (per user, per channel, per guild, etc.).
        bypass_roles (Iterable[int]): An iterable of role IDs that bypass the cooldown.

    Returns:
        Callable: A decorator that adds the described cooldown behavior to the command.

    Raises:
        TypeError: If the decorator is not applied to an instance of `Command`.
    """
    # Make it a set so lookup is hash based.
    bypass = set(bypass_roles)

    # This handles the actual cooldown logic.
    buckets = CooldownMapping(Cooldown(rate, per, type))

    # Will be called after the command has been parse but before it has been invoked, ensures that
    # the cooldown won't be updated if the user screws up their input to the command.
    async def predicate(cog: Cog, ctx: Context) -> None:
        nonlocal bypass, buckets

        if any(role.id in bypass for role in ctx.author.roles):
            return

        # Cooldown logic, taken from discord.py internals.
        current = ctx.message.created_at.replace(tzinfo=datetime.UTC).timestamp()
        bucket = buckets.get_bucket(ctx.message)
        retry_after = bucket.update_rate_limit(current)
        if retry_after:
            raise CommandOnCooldown(bucket, retry_after)

    def wrapper(command: Command) -> Command:
        # NOTE: this could be changed if a subclass of Command were to be used. I didn't see the need for it
        # so I just made it raise an error when the decorator is applied before the actual command object exists.
        #
        # If the `before_invoke` detail is ever a problem then I can quickly just swap over.
        if not isinstance(command, Command):
            raise TypeError(
                "Decorator `cooldown_with_role_bypass` must be applied after the command decorator. "
                "This means it has to be above the command decorator in the code."
            )

        command._before_invoke = predicate

        return command

    return wrapper


async def has_any_role_check(ctx: Context, *roles: str | int) -> bool:
    """
    Verify if the context's author has any of the specified roles.

    This check will always fail if the context is a DM, since DMs don't have roles.

    Args:
        ctx (:obj:`discord.ext.commands.Context`): The context where the command is being invoked.
        roles (Union[str, int], ...): A collection of role IDs.

    Returns:
        bool: True if the context's author has at least one of the specified roles; False otherwise.
    """
    if not ctx.guild:  # Return False in a DM
        log.trace(
            f"{ctx.author} tried to use the '{ctx.command.name}'command from a DM. "
            "This command is restricted by the with_role decorator. Rejecting request."
        )
        return False

    for role in ctx.author.roles:
        if role.id in roles:
            log.trace(f"{ctx.author} has the '{role.name}' role, and passes the check.")
            return True

    log.trace(
        f"{ctx.author} does not have the required role to use "
        f"the '{ctx.command.name}' command, so the request is rejected."
    )
    return False


async def has_no_roles_check(ctx: Context, *roles: str | int) -> bool:
    """
    Verify if the context's author doesn't have any of the specified roles.

    This check will always fail if the context is a DM, since DMs don't have roles.

    Args:
        ctx (:obj:`discord.ext.commands.Context`): The context where the command is being invoked.
        roles (Union[str, int], ...): A collection of role IDs.

    Returns:
        bool: True if the context's author doesn't have any of the specified roles; False otherwise.
    """
    if not ctx.guild:  # Return False in a DM
        log.trace(
            f"{ctx.author} tried to use the '{ctx.command.name}' command from a DM. "
            "This command is restricted by the without_role decorator. Rejecting request."
        )
        return False

    author_roles = [role.id for role in ctx.author.roles]
    check = all(role not in author_roles for role in roles)
    log.trace(
        f"{ctx.author} tried to call the '{ctx.command.name}' command. "
        f"The result of the without_role check was {check}."
    )
    return check



"""Common regular expressions."""

import re

DISCORD_INVITE = re.compile(
    r"(https?://)?(www\.)?"                      # Optional http(s) and www.
    r"(discord([.,]|dot)gg|"                     # Could be discord.gg/
    r"discord([.,]|dot)com(/|slash)invite|"      # or discord.com/invite/
    r"discordapp([.,]|dot)com(/|slash)invite|"   # or discordapp.com/invite/
    r"discord([.,]|dot)me|"                      # or discord.me
    r"discord([.,]|dot)li|"                      # or discord.li
    r"discord([.,]|dot)io|"                      # or discord.io.
    r"((?<!\w)([.,]|dot))gg"                     # or .gg/
    r")(/|slash)"                                # / or 'slash'
    r"(?P<invite>\S+)",                          # the invite code itself
    flags=re.IGNORECASE
)
"""
Regex for Discord server invites.

.. warning::
    This regex pattern will capture until a whitespace, if you are to use the 'invite' capture group in
    any HTTP requests or similar. Please ensure you sanitise the output using something
    such as :func:`urllib.parse.quote`.

:meta hide-value:
"""

FORMATTED_CODE_REGEX = re.compile(
    r"(?P<delim>(?P<block>```)|``?)"        # code delimiter: 1-3 backticks; (?P=block) only matches if it's a block
    r"(?(block)(?:(?P<lang>[a-z]+)\n)?)"    # if we're in a block, match optional language (only letters plus newline)
    r"(?:[ \t]*\n)*"                        # any blank (empty or tabs/spaces only) lines before the code
    r"(?P<code>.*?)"                        # extract all code inside the markup
    r"\s*"                                  # any more whitespace before the end of the code markup
    r"(?P=delim)",                          # match the exact same delimiter from the start again
    flags=re.DOTALL | re.IGNORECASE         # "." also matches newlines, case insensitive
)
"""
Regex for formatted code, using Discord's code blocks.

:meta hide-value:
"""

RAW_CODE_REGEX = re.compile(
    r"^(?:[ \t]*\n)*"                       # any blank (empty or tabs/spaces only) lines before the code
    r"(?P<code>.*?)"                        # extract all the rest as code
    r"\s*$",                                # any trailing whitespace until the end of the string
    flags=re.DOTALL                         # "." also matches newlines
)
"""
Regex for raw code, *not* using Discord's code blocks.

:meta hide-value:
"""



"""Useful utilities and tools for Discord bot development."""

from pydis_core import async_stats, exts, site_api, utils
from pydis_core._bot import BotBase, StartupError
from pydis_core.utils.pagination import EmptyPaginatorEmbedError, LinePaginator, PaginationEmojis

__all__ = [
    async_stats,
    BotBase,
    exts,
    utils,
    site_api,
    StartupError,
    LinePaginator,
    PaginationEmojis,
    EmptyPaginatorEmbedError
]

__all__ = [module.__name__ for module in __all__]



from collections.abc import Sequence

import discord

from pydis_core.utils.logging import get_logger
from pydis_core.utils.scheduling import create_task

log = get_logger(__name__)


def reaction_check(
    reaction: discord.Reaction,
    user: discord.abc.User,
    *,
    message_id: int,
    allowed_emoji: Sequence[str],
    allowed_users: Sequence[int],
    allowed_roles: Sequence[int] | None = None,
) -> bool:
    """
    Check if a reaction's emoji and author are allowed and the message is `message_id`.

    If the user is not allowed, remove the reaction. Ignore reactions made by the bot.
    If `allow_mods` is True, allow users with moderator roles even if they're not in `allowed_users`.
    """
    right_reaction = (
        not user.bot
        and reaction.message.id == message_id
        and str(reaction.emoji) in allowed_emoji
    )
    if not right_reaction:
        return False

    allowed_roles = allowed_roles or []
    has_sufficient_roles = any(role.id in allowed_roles for role in getattr(user, "roles", []))

    if user.id in allowed_users or has_sufficient_roles:
        log.trace(f"Allowed reaction {reaction} by {user} on {reaction.message.id}.")
        return True

    log.trace(f"Removing reaction {reaction} by {user} on {reaction.message.id}: disallowed user.")
    create_task(
        reaction.message.remove_reaction(reaction.emoji, user),
        suppressed_exceptions=(discord.HTTPException,),
        name=f"remove_reaction-{reaction}-{reaction.message.id}-{user}"
    )
    return False



"""Useful helper functions for interactin with :obj:`discord.Member` objects."""
import typing
from collections import abc

import discord

from pydis_core.utils import logging

log = logging.get_logger(__name__)


async def get_or_fetch_member(guild: discord.Guild, member_id: int) -> discord.Member | None:
    """
    Attempt to get a member from cache; on failure fetch from the API.

    Returns:
        The :obj:`discord.Member` or :obj:`None` to indicate the member could not be found.
    """
    if member := guild.get_member(member_id):
        log.trace(f"{member} retrieved from cache.")
        return member
    try:
        member = await guild.fetch_member(member_id)
    except discord.errors.HTTPException as e:
        log.trace(f"Failed to fetch {member_id} from API.")
        if e.status in [400, 404]:
            return None
        raise
    log.trace(f"{member} fetched from API.")
    return member


async def handle_role_change(
    member: discord.Member,
    coro: typing.Callable[[discord.Role], abc.Coroutine],
    role: discord.Role
) -> None:
    """
    Await the given ``coro`` with ``role`` as the sole argument.

    Handle errors that we expect to be raised from
    :obj:`discord.Member.add_roles` and :obj:`discord.Member.remove_roles`.

    Args:
        member: The member that is being modified for logging purposes.
        coro: This is intended to be :obj:`discord.Member.add_roles` or :obj:`discord.Member.remove_roles`.
        role: The role to be passed to ``coro``.
    """
    try:
        await coro(role)
    except discord.NotFound:
        log.error(f"Failed to change role for {member} ({member.id}): member not found")
    except discord.Forbidden:
        log.error(
            f"Forbidden to change role for {member} ({member.id}); "
            f"possibly due to role hierarchy"
        )
    except discord.HTTPException as e:
        log.error(f"Failed to change role for {member} ({member.id}): {e.status} {e.code}")
